class NoPackageManagerException(Exception):
    pass


class MissingPackageException(Exception):
    pass


class InvalidScanModeException(Exception):
    pass


class ScanFailedException(Exception):
    pass


class MaliciousPackageException(Exception):
    pass


class MissingAPIKeyException(Exception):
    pass


class MissingSBOMException(Exception):
    pass


class ScanTimeoutException(Exception):
    pass



from __future__ import annotations
import json
import logging
import requests
import time

from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
from typing import Tuple, Final

from ossprey.exceptions import (
    MissingAPIKeyException,
    MissingSBOMException,
    ScanFailedException,
    ScanTimeoutException,
)

logger = logging.getLogger(__name__)


class Ossprey:

    def __init__(self, api_url: str, api_key: str):
        self.api_url = api_url
        self.api_key = api_key

        if not self.api_key:
            raise MissingAPIKeyException("API Key is null or empty")

        self.session = self.create_session()

    # This takes a python dictionary and submits it to the API
    def validate(self, minibom: dict) -> dict | None:

        response = self.submit(minibom)

        match response.status_code:
            case 200:
                return response.json()
            case 202:
                json = response.json()
                sbom_id = json["sbom_id"]
                scan_id = json["scan_id"]
                return self.wait_for_completion(sbom_id, scan_id)
            case 429:
                logger.error("Rate limit exceeded")
                return None
            case _:
                logger.error("Failed to submit request")
                logger.debug(f"Status code: {response.status_code}")
                logger.debug(f"Response: {response.text}")

                data = response.json()
                if "message" in data:
                    logger.error(data["message"])
                return None

    def submit(self, json_bom: dict) -> requests.Response:

        # Get the url
        url = self.api_url + "/public/v1/scans"

        logger.debug(f"JSON Submission: {json.dumps(json_bom)}")

        # Submit bom to API
        json_data = {"sbom": json_bom}
        headers = {"Content-Type": "application/json", "x-api-key": self.api_key}
        response = self.session.post(url, headers=headers, json=json_data)

        return response

    def wait_for_completion(self, sbom_id: str, scan_id: str) -> dict:
        url = self.api_url + "/public/v1/scans/status"

        params = {"sbom_id": sbom_id, "scan_id": scan_id}

        headers = {"Content-Type": "application/json", "x-api-key": self.api_key}
        for i in range(1, 20):
            time.sleep(i * i)

            response = self.session.get(url, headers=headers, params=params)
            if response.status_code not in [200, 202]:
                logger.error("Error returned when retrieving the results")
                logger.debug(f"Status code: {response.status_code}")
                logger.debug(f"Response: {response.text}")
                raise ScanFailedException("Error returned when retrieving the results")

            ret = response.json()
            if ret["status"] == "SUCCEEDED":
                if "output" in ret:
                    return ret["output"]
                else:
                    raise MissingSBOMException("Error no SBOM returned")

        logger.error("Scan took too long to complete")
        raise ScanTimeoutException("Scan took too long to complete")

    @staticmethod
    def create_session() -> requests.Session:
        """Return a Session that transparently retries on 503."""

        _RETRIES: Final = 5
        _BACKOFF: Final = 1.0  # seconds
        _STATUS_FORCELIST = (503,)
        _ALLOWED_METHODS: Final[Tuple[str, ...]] = (
            "GET",
            "POST",
        )

        retry = Retry(
            total=_RETRIES,
            backoff_factor=_BACKOFF,
            status_forcelist=_STATUS_FORCELIST,
            allowed_methods=_ALLOWED_METHODS,
            respect_retry_after_header=True,
        )
        adapter = HTTPAdapter(max_retries=retry)
        sess = requests.Session()
        sess.mount("http://", adapter)
        sess.mount("https://", adapter)
        return sess



from __future__ import annotations
import json
import logging
import os
from subprocess import CalledProcessError

from ossprey.environment import get_environment_details
from ossprey.exceptions import (
    InvalidScanModeException,
    MissingPackageException,
    NoPackageManagerException,
    ScanFailedException,
)
from ossprey.modes import get_modes, get_all_modes
from ossprey.sbom_python import (
    update_sbom_from_requirements,
    update_sbom_from_poetry,
    update_sbom_from_virtualenv,
)
from ossprey.sbom_javascript import update_sbom_from_npm, update_sbom_from_yarn
from ossprey.sbom_filesystem import update_sbom_from_filesystem
from ossprey.ossprey import Ossprey

from ossbom.converters.factory import SBOMConverterFactory
from ossbom.model.ossbom import OSSBOM
from ossbom.model.vulnerability import Vulnerability

logger = logging.getLogger(__name__)


def scan_python(modes: list[str], sbom: OSSBOM, package_name: str) -> OSSBOM:

    if "python-requirements" in modes:
        sbom = update_sbom_from_requirements(sbom, package_name + "/requirements.txt")

    if "poetry" in modes:
        try:
            sbom = update_sbom_from_poetry(sbom, package_name)
        except CalledProcessError as e:
            logger.error(
                "Error running poetry scan, skipping poetry scan, attempting pipenv scan instead"
            )
            logger.debug(e.stderr)
            logger.debug("--")
            logger.debug(e.stdout)

            # Appending pipenv to modes to attempt to get some python dependencies if poetry fails
            modes.append("pipenv")

    if "pipenv" in modes:
        sbom = update_sbom_from_virtualenv(sbom, package_name)

    return sbom


def scan_javascript(modes: list[str], sbom: OSSBOM, package_name: str) -> OSSBOM:
    if "npm" in modes:
        sbom = update_sbom_from_npm(sbom, package_name)

    if "yarn" in modes:
        sbom = update_sbom_from_yarn(sbom, package_name)

    return sbom


def scan_filesystem(modes: list[str], sbom: OSSBOM, package_name: str) -> OSSBOM:

    if "fs" in modes:
        sbom = update_sbom_from_filesystem(sbom, package_name)

    return sbom


def scan(
    package_name: str,
    mode: str = "auto",
    local_scan: str | None = None,
    url: str | None = None,
    api_key: str | None = None,
) -> OSSBOM:

    if mode == "auto":
        logger.debug("Auto mode selected")

        # Check the folder for files that map to different package managers
        modes = get_modes(package_name)
        if len(modes) == 0:
            logger.error("No package manager found")
            raise NoPackageManagerException("No package manager found in the directory")
    else:
        modes = [mode]

    logger.info(f"Scanning {package_name} with modes: {modes}")
    # If package location doesn't exist, raise an error
    if not os.path.exists(package_name):
        logger.error(f"Package {package_name} does not exist")
        raise MissingPackageException(f"Package {package_name} does not exist")

    sbom = OSSBOM()

    if any(mode not in get_all_modes() for mode in modes) or len(modes) == 0:
        raise InvalidScanModeException("Invalid scanning method: " + str(modes))

    # Enrich the sbom based on the scans provided
    sbom = scan_python(modes, sbom, package_name)
    sbom = scan_javascript(modes, sbom, package_name)
    sbom = scan_filesystem(modes, sbom, package_name)

    # Update sbom to contain the local environment
    env = get_environment_details(package_name)
    sbom.update_environment(env)

    logger.info(f"Scanning {len(sbom.get_components())} components")

    if not local_scan:
        ossprey = Ossprey(url, api_key)

        # Compress to MINIBOM
        sbom = SBOMConverterFactory.to_minibom(sbom)

        sbom = ossprey.validate(sbom)
        if not sbom:
            raise ScanFailedException("Issue OSSPREY Service")

        # Convert to OSSBOM
        sbom = SBOMConverterFactory.from_minibom(sbom)

    if local_scan == "dry-run-malicious":
        # Add a vulnerability for testing purposes
        components = sbom.get_components()
        if len(components) == 0:
            raise Exception("No components found to add a test vulnerability")

        component = sbom.get_components()[0]
        purl = f"pkg:{component.type}/{component.name}@{component.version}"
        vulnerability = Vulnerability(
            id="TEST-2024-0001",
            purl=purl,
            description="This is a test vulnerability added in dry-run-malicious mode",
        )
        sbom.add_vulnerability(vulnerability)

    logger.debug(json.dumps(sbom.to_dict(), indent=4))

    return sbom
